function [xc,histout,costdata,fail] = geom_cg(prob, opts, x0)
%
% Adapted from steep by C. T. Kelley, Dec 20, 1996
%
% function [x,histout,costdata] = steep(x0,f,tol,maxit)
%
% Geometric cg descent with Armijo rule, polynomial linesearch 
% 
%
% Input: x0 = initial iterate
%        f = objective function,
%            the calling sequence for f should be
%            [fout,gout]=f(x) where fout=f(x) is a scalar
%              and gout = grad f(x) is a COLUMN vector
%        tol = termination criterion norm(grad) < tol
%              optional, default = 1.d-6
%        maxit = maximum iterations (optional) default = 1000
%
% Output: x = solution
%         histout = iteration history   
%             Each row of histout is
%            [norm(grad), f, number of step length reductions, orthog dirs, restarts]
%
% Requires: polymod.m
%
% linesearch parms
% 


t_begin = tic();

TOL_RESCHG = 1e-8;

%beta_type = 'F-R';
beta_type = 'P-R';
%beta_type = 'subspace'; -> does not work so well

fail = true;

norm_M_Omega = norm(prob.data);

ORTH_VALUE = 0.1; % the search directions should be almost orthogonal

itc = 1; xc = x0; 
fc = F(prob,xc);
gc = grad(prob,xc);
ip_gc = ip(xc,gc,gc);
fold = 2*fc;
% first search-dir is steepest gradient
dir = scaleTxM(gc,-1);

numf = 1; numg = 1; numh = 0;
ithist=zeros(opts.maxit,5);
ithist(1,1)=sqrt(ip_gc); ithist(1,2) = fc; ithist(1,4)=itc-1; ithist(1,3)=0; 
xc_old1 = xc;
beta = 0;


for itc=1:opts.maxit   
        
    if strcmp(beta_type, 'subspace')
        % -- dir1 = transport(xc-xc_old1) to a tangent vector at xc;
        Sk = diag(xc.sigma);
        Sk1 = diag(xc_old1.sigma);
        dir1.M = Sk - (xc.U'*xc_old1.U)*Sk1*(xc_old1.V'*xc.V);
        dir1.Up = - xc_old1.U*(Sk1*(xc_old1.V'*xc.V)) + ...
            xc.U*((xc.U'*xc_old1.U)*Sk1*(xc_old1.V'*xc.V));        
        dir1.Vp = - xc_old1.V*(Sk1*(xc_old1.U'*xc.U)) + ...
            xc.V*((xc.V'*xc_old1.V)*Sk1*(xc_old1.U'*xc.U));
        
        
        
%         % -- dir2 = xc-xcold2;
%         Sk = diag(xc.sigma);
%         Sk2 = diag(xc_old2.sigma);
%         dir2.M = Sk - (xc.U'*xc_old2.U)*Sk2*(xc_old2.V'*xc.V);
%         dir2.Up = - xc_old2.U*(Sk2*(xc_old2.V'*xc.V)) + ...
%             xc.U*((xc.U'*xc_old2.U)*Sk2*(xc_old2.V'*xc.V));
%         
%         dir2.Vp = - xc_old2.V*(Sk2*(xc_old2.U'*xc.U)) + ...
%             xc.V*((xc.V'*xc_old2.V)*Sk2*(xc_old2.U'*xc.U));
%         
%         % -- dir3 = xc-xcold3;
%         Sk = diag(xc.sigma);
%         Sk3 = diag(xc_old3.sigma);
%         dir3.M = Sk - (xc.U'*xc_old3.U)*Sk3*(xc_old3.V'*xc.V);
%         dir3.Up = - xc_old3.U*(Sk3*(xc_old3.V'*xc.V)) + ...
%             xc.U*((xc.U'*xc_old3.U)*Sk3*(xc_old3.V'*xc.V));
%         
%         dir3.Vp = - xc_old3.V*(Sk3*(xc_old3.U'*xc.U)) + ...
%             xc.V*((xc.V'*xc_old3.V)*Sk3*(xc_old3.U'*xc.U));
        
            omegas = exact_search_onlyTxM_multiple(prob, xc, {gc dir1});
            dir = plusTxM(gc,dir1,omegas(1),omegas(2)); 
            tinit = 1;
            [xc_new,fc,succ,numf_a,iarm] = armijo_search(prob, xc,fc,dir);
        
    else % else do a 1d search
        %if beta==0
        %    %exact_search_onR1(prob, xc,dir)
            tinit(itc) = exact_search_onR1(prob, xc,dir);
        %else
        %    tinit(itc) = exact_search_onlyTxM_multiple(prob, xc, {dir});
        %end
        %[xc_new,fc,succ,numf_a,iarm, tinit(itc)] = exact_search(prob, xc,fc,dir,tinit(itc),tinit(itc)*5);
        fc_b = fc;
        %tinit_itc = tinit(itc)
        [xc_new,fc,succ,numf_a,iarm,tinit(itc)] = armijo_search(prob, xc,fc,dir, tinit(itc),false);
    end    
    
    if ~succ && beta ~= 0
        if opts.verbosity > 0; warning('Line search failed on CG. Resetting to gradient.'); end
        beta = 0;
        % if we did cg, reset to steepest descent and try again
        dir = scaleTxM(gc,-1);
        tinit(itc) = exact_search_onR1(prob, xc,dir);
        %tinit(itc) = exact_search_onlyTxM_multiple(prob, xc, {dir});
        [xc_new,fc,succ,numf_a,iarm,tinit(itc)] = armijo_search(prob, xc,fc_b,dir, tinit(itc));
    end
        
    % if it still fails (beta is always 0 here -> steepest descent)
    % then we give up
    if ~succ   
        iarm
        beta
        xc = xc_new;   
        ithist(itc,1)=sqrt(ip_gc_new)/max(1,norm(xc.sigma)); ithist(itc,2) = sqrt(2*fc);
        ithist(itc,3)=beta; ithist(itc,5) = toc(t_begin);
        ithist(itc,4) = reschg;
        histout=ithist(1:itc,:); costdata=[numf, numg, numh];
        if opts.verbosity > 0; warning('Line search failed on steepest descent. Exiting...'); end
        return
    end    
    
    
    % grad(new x)
    gc_new = grad(prob,xc_new);
    ip_gc_new = ip(xc_new,gc_new,gc_new);
    
    % Test for convergence
    if sqrt(2*fc) < opts.abs_f_tol
        if opts.verbosity > 0
            disp('Abs f tol reached.')
        end
        fail = false;
        break;
    end
    if sqrt(2*fc)/norm_M_Omega < opts.rel_f_tol
        if opts.verbosity > 0; warning('Relative f tol reached.'); end
        fail = false;
        break;
    end
    
    rel_grad = sqrt(ip_gc)/max(1,norm(xc_new.sigma));
    if rel_grad < opts.rel_grad_tol 
        if opts.verbosity > 0; warning('Relative gradient tol reached.'); end
        fail = false;
        break;
    end
    
    % for 'stagnation stopping criterion' after 10 iters
    reschg = abs(1-sqrt(2*fc)/sqrt(2*fold) );  % LMARank's detection
    %reschg = abs(sqrt(2*fc) - sqrt(2*fold)) / max(1,norm_M_Omega);
    if itc > 10 && reschg < TOL_RESCHG
        if opts.verbosity > 0; warning('Iteration stagnated.'); end
        fail = true;
        break;
    end
    
      
    
    % new dir = -grad(new x) 
    %           + beta * vectTransp(old x, old dir, tmin*old dir)
    
    %gc_old = transpVect2(prob,xc,gc, tinit(itc),dir); 
    gc_old = transpVect(prob,xc,gc,xc_new,1);
    
    %dir = transpVect2(prob,xc,dir, tinit(itc),dir);
    dir = transpVect(prob,xc,dir,xc_new,1);
    
    
    
    % we test how orthogonal the previous gradient is with
    % the current gradient, and possibly reset the to the gradient
    orth_grads = ip(xc_new,gc_old,gc_new)/ip_gc_new;
    
    if (orth_grads) >= ORTH_VALUE
        if opts.verbosity > 0; warning('New gradient is almost orthogonal to current gradient. This is good, so we can reset to steepest descent.'); end
        beta = 0;
        dir = plusTxM(gc_new, dir, -1, beta);
        
    else % Compute the CG modification    
        % beta
        if strcmp(beta_type, 'F-R')  % Fletcher-Reeves
            beta = ip_gc_new / ip_gc;
            % new dir
            dir = plusTxM(gc_new, dir, -1, beta);
        elseif strcmp(beta_type, 'P-R')  % Polak-Ribiere
            % vector grad(new) - transported grad(current)
            diff = plusTxM(gc_new, gc_old, 1, -1);
            ip_diff = ip(xc_new, gc_new, diff);
            beta = ip_diff / ip_gc;
            
            %dk_c = transpVect(prob, xc,dir,xc_new,1);
            %ct = -1/(sqrt(ip(xc_new,dk_c,dk_c))*min(0.01,sqrt(ip_gc)));
            %beta = max(ct,beta);
            beta = max(0,beta);
            dir = plusTxM(gc_new, dir, -1, beta);
            
            %%%CHANGE THIS
            %dir = scaleTxM(gc_new,-1);
            
        elseif strcmp(beta_type, 'H-Z')  % Hager-Zhang
            yk_c = plusTxM(gc_new, gc_old, 1, -1);
            dk_c = transpVect(prob, xc,dir,xc_new,1);
            temp = plusTxM(yk_c,dk_c,1,-2*ip(xc_new,yk_c,yk_c)/ip(xc_new,yk_c,dk_c));
            beta = 1/ip(xc_new,yk_c,dk_c)*ip(xc_new,temp,gc_new);
            
            ct = -1/(sqrt(ip(xc_new,dk_c,dk_c))*min(0.01,sqrt(ip_gc)));
            beta = max(ct,beta);
            % new dir
            dir = plusTxM(gc_new, dir, -1, beta);
        elseif strcmp(beta_type, 'subspace')
            %if restart_subspace
            %    dir = gc_new;
            %end
            
            %dir2 = transpVect(prob,xc,dir1,xc_new,1);
            %dir1 = dir;
        end
    end
    
    % check if dir is descent, if not take -gradient (i.e. beta=0)
    if ~strcmp(beta_type, 'subspace')
        g0 = ip(xc_new,gc_new,dir);
        if g0>=0
            warning('New search direction not a descent direction. Resetting to -grad.')
            dir = scaleTxM(gc_new, -1);
            beta = 0;
        end
    end
    
    
    % update _new to current
    gc = gc_new;
    ip_gc = ip_gc_new;
    xc_old1 = xc;
    xc = xc_new;
    fold = fc;

    numg = numg+1;
    
	ithist(itc,1) = rel_grad;
    ithist(itc,2) = sqrt(2*fc); 
	ithist(itc,4) = reschg; 
    ithist(itc,3) = beta;
    ithist(itc,5) = toc(t_begin);
   
end

xc = xc_new;   
ithist(itc,1)=sqrt(ip_gc_new)/max(1,norm(xc.sigma)); ithist(itc,2) = sqrt(2*fc);
ithist(itc,3)=beta; ithist(itc,5) = toc(t_begin);
ithist(itc,4) = reschg;
histout=ithist(1:itc,:); costdata=[numf, numg, numh];
