function varargout = rtr(fns,params,prob,opts)
% RTR   Riemannian Trust-Region (with tCG inner solve)
%
% This performs a retraction-based, trust-region minimization of an objective function on a 
% Riemannian manifold.
%
% x = rtr(fns,params) returns the minimizer (a point on the manifold)
% [x,stats] = rtr(fns,params) also returns some statistics from the algorithm
%
% The struct fns contains the function handles necessary to perform the optimization:
%  fns.R(x,eta) : retract tangent vector eta at T_x M to the manifold
%  fns.g(x,eta,zeta) : Riemannian metric at T_x M 
%  fns.proj(x,eta) : Project eta from nearby T_x M to T_x M (used in iteration to combat round-off)
%                    Also used in debugging to check tangentiality
%                    Can be set to the identity if this is not a concern.
%  fns.f(x) : Compute the objective function at x
%  fns.fgrad(x) : Compute the gradient of f at x, returning a tangent vector
%  fns.fhess(x,eta) : Apply the Hessian of f, hess_x: T_x M -> T_x M
%  fns.scale(h,c) : scale a tangent vector with a real number c
%  fns.plus(h1,h2,c1,c2) : add two (scaled) tangent vectors
% Optionally, three other methods may be specified:
%  fns.prec(x,eta) : apply an s.p.d. preconditioner for the inner iteration, to eta, a
%                    tangent vector in T_x M
%  fns.dist(x,y) : returns the distance on the manifold between points x and y
%  fns.randT(x) : returns a very small random tangentt vector in T_x M
% More info:
%  fns.dist is used to measure the distance from solution and is only referenced if params.xsol was specified.
%  fns.randT is used to initialize the trust-region subproblem with a random vector, to encourage escape from an
%      exact critical point.
%  fns.prec is a preconditioner for the model minimization, a positive-definite (undr fns.g)
%      mapping from T_x M to T_x M
%
% Parameters to the solver include:
% Required parameters:
%   params.x0        - An initial iterate. Because RTR does not know what the manifold is, this is mandatory.
%   params.Delta_bar - Maximum trust-region radius. Because RTR does not know anything about the manifold, this is mandatory.
%   params.Delta0    - Initial trust-region radius. Because RTR does not know anything about the manifold, this is mandatory.
%   params.xsol      - For testing distance from solution using fns.dist
%   params.verbosity - Level of verbosity: 0 is silent, 1 has one-line info, 2 has detailed info
%   params.debug     - Debugging tests (0 is silent, 1 has some, 2 has a lot, 3 is more than you want to know)
%   params.min_outer - Minimum number of outer iterations (default: 0); used only with randomization
%   params.max_outer - Maximum number of outer iterations (default: 100)
%   params.min_inner - Minimum number of inner iterations (default: 0). Only in effect if using randomized initial tangent vectors.
%   params.max_inner - Maximum number of inner iterations (default: inf). Recommended: dimension of manifold.
%   params.epsilon   - Outer Convergence tolerance (absolute)
%   params.kappa     - Inner kappa convergence tolerance
%   params.theta     - Inner theta convergence tolerance
%   params.rho_prime - Accept/reject ratio
%   params.testgh    - perform some simple numerical testing of gradient and Hessian
%   params.useRand   - initial the trust-region solve with a random tangent vector
%
% The stats output is a struct array, with fields:
%  k  - the outer iteration number for the stats
%  ng - the norm of the gradient, sqrt(g(x,gradfx,gradfx))
%  fx - the current value under the objective function
%  rho - the performance ratio for the iterate
%  time - the wallclock time for the outer iteration
%  accepted - whether the proposed iterate was accepted or not
%  numinner - the number of inner iterations used to compute the next iterate
%  Delta - the trust-region radius at the outer iteration
%  dist     - the distance from the solution
%  cauchy   - whether the cauchy point was used in place of an updated computed from a random initial tangent vector
%
% See also irtr

% About: RTR - Riemannian Trust-Region
% (C) 2004-2007, P.-A. Absil, C. G. Baker, K. A. Gallivan
% Florida State University
% School of Computational Science

% Modification history
% Version 0.221 - CGB
%   Using preconditioner-based trust-region radius
%   Disable randomization when using preconditioner
% Version 0.2 - CGB
%   Added more commenting
%   Added preconditioning for inner iteration (fns.prec)
% Version 0.1 - CGB - Sat Nov 04 2006
%   Started from RTR03(Baker), from tCG07(Absil)

tcg_stop_reason = {'negative curvature',...
                   'exceeded trust region',...
                   'linear convergence',...
                   'super-linear convergence',...
                   'dimension exceeded'};
lan_stop_reason = {'reached residual (interior)',...
                   'reached residual (exterior)',...
                   'breakdown - reduced T',...
                   'dimension exceeded'};
norm_M_Omega = norm(prob.data);
shrunk = false;

if nargin < 2,
   error('Invalid arguments to rtr. See ''help rtr''');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% check functions handles
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if   ~(isfield(fns,'R')     && isa(fcnchk(fns.R)    ,'function_handle')) ...
  || ~(isfield(fns,'g')     && isa(fcnchk(fns.g)    ,'function_handle')) ...
  || ~(isfield(fns,'proj')  && isa(fcnchk(fns.proj) ,'function_handle')) ...
  || ~(isfield(fns,'f')     && isa(fcnchk(fns.f)    ,'function_handle')) ...
  || ~(isfield(fns,'fgrad') && isa(fcnchk(fns.fgrad),'function_handle')) ...
  || ~(isfield(fns,'fhess') && isa(fcnchk(fns.fhess),'function_handle')),
  error('Invalid arguments to rtr. See ''help rtr''');
end
fns.havedist = 0;
fns.haverand = 0;
fns.haveprec = 0;
if isfield(fns,'dist') && isa(fcnchk(fns.dist),'function_handle')
    fns.havedist = 1;
end
if isfield(fns,'randT') && isa(fcnchk(fns.randT),'function_handle')
    fns.haverand = 1;
else
    fns.haverand = 0;
end
if isfield(fns,'prec') && isa(fcnchk(fns.prec),'function_handle'),
    fns.haveprec = 1;
else
    fns.haveprec = 0;
end
%%EDIT%%
if isfield(fns,'scale') && isa(fcnchk(fns.scale),'function_handle'),
    fns.havescale = 1;
else
    fns.havescale = 0;
end
if isfield(fns,'plus') && isa(fcnchk(fns.scale),'function_handle'),
    fns.haveplus = 1;
else
    fns.haveplus = 0;
end
if isfield(fns,'magic') && isa(fcnchk(fns.magic),'function_handle'),
    fns.havemagic = 1;
else
    fns.havemagic = 0;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% set default parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
verbosity =   1;
debug     =   0;
min_inner =   0;
max_inner = inf;
min_outer =   3;
max_outer = 100;
epsilon   =   1e-6;
kappa     =   0.1;
theta     =   0.8;
rho_prime =   0.1;
kappa_easy = .001;
testgh    = 0;
useRand  = 0;
solver   = 0; % tCG
magic = 0;
%Delta_bar =  user must specify
%Delta0    =  user must specify
%x0        =  user must specify

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% get algorithm parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[verbosity,params] = get_int(params,'verbosity','Level of verbosity'                ,verbosity);
[debug    ,params] = get_int(params,'debug'    ,'Debugging tests'                   ,debug);
[min_outer,params] = get_int(params,'min_outer','Minimum number of outer iterations',min_outer,0);
[max_outer,params] = get_int(params,'max_outer','Maximum number of outer iterations',max_outer,0);
[min_inner,params] = get_int(params,'min_inner','Minimum number of inner iterations',min_inner,0);
[max_inner,params] = get_int(params,'max_inner','Maximum number of inner iterations',max_inner,0);
[testgh   ,params] = get_int(params,'testgh'   ,'Test gradient and Hessian'         ,testgh);
[useRand  ,params]  = get_int(params,'useRand'  ,'Random TR subproblem init'        ,useRand);
[solver   ,params]  = get_int(params,'solver'  ,'TR subproblem solver'              ,solver);
[x0       ,params] = get_iterate(params,'x0'   ,'Initial iterate');
[xsol     ,params] = get_iterate(params,'xsol' ,'Solution for testing',[]);
if isempty(xsol),
    % must have fns.dist and xsol to make fns.dist() comparisons
    fns.havedist = 0;
end
[epsilon  ,params] = get_float(params,'epsilon'  ,'Outer Convergence tolerance'      ,epsilon,0);
[rho_prime,params] = get_float(params,'rho_prime','Accept/reject parameter'          ,rho_prime,0,1);
[Delta_bar,params] = get_float(params,'Delta_bar','Maximum trust-region radius');
[Delta0   ,params] = get_float(params,'Delta0'   ,'Initial trust-region radius');
[kappa    ,params] = get_float(params,'kappa'    ,'Inner kappa convergence tolerance',kappa,0,1);
[theta    ,params] = get_float(params,'theta'    ,'Inner theta convergence tolerance',theta,0);
[kappa_easy,params] = get_float(params,'kappa_easy','Inner convergence tolerance'    ,kappa_easy,0,1);


if testgh && fns.haverand,
   TestGH(x0,fns);
   varargout{1} = x0;
   if nargout > 1,
      varargout{2} = struct([]);
   end
   return;
end

if fns.haveprec && useRand && fns.haverand,
   fprintf('rtr/tCG: Cannot perform preconditioned tCG with random initial vector.\n');
   fprintf('rtr/tCG: Disabling randomization.\n');
   useRand = 0;
end

if ~fns.haverand,
   useRand = 0;
end

%verbosity = 2; % temp

% ***** Initializations *****

% initialize counters/sentinals
% allocate storage for dist, counters
k          = 0;  % counter for outer (TR) iteration.
stop_outer = 0;  % stopping criterion for TR.
total_time = 0;     % total time spent in outer loop


% initialize solution and companion measures: 
% fgrad(x)
% f(x)
% tic

x = x0;
x = prepx(prob, x);

fx = fns.f(x);
fgradx = fns.fgrad(x);
norm_grad = sqrt(fns.g(x,fgradx,fgradx));
% precondition the residual
%if fns.haveprec == 0,
    z = fgradx;
%else
%    z = fns.prec(x,fgradx);
%end
%crit_measure = sqrt(fns.g(x,fgradx,z));
crit_measure = norm_grad;
% this_time = toc;

% iter time
iter_time = zeros(max_outer+1,1);
iter_time(1) = 0; 
costs = zeros(max_outer+1,1);
costs(1) = fx;
infos.numit = [0];
infos.hessit = [0];

% initialize trust-region radius
Delta = Delta0;

% ** Record data:
curstat.k        = k;
curstat.ng       = norm_grad;
curstat.crit     = crit_measure;
curstat.fx       = fx;
curstat.rho      = inf;
curstat.rhonum      = 0;
curstat.rhoden      = 0;
%curstat.time = this_time;
curstat.accepted = true;
curstat.numinner = 0;
curstat.Delta    = Delta;
if fns.havedist,
    curstat.dist = fns.dist(x,xsol);
end
if useRand,
    curstat.cauchy = false;
end
stats(1) = curstat;

% ** Display:
if (verbosity == 1)
   fprintf(['%3s %3s      %5s                %5s     ',...
            'cost: %e f: %e   |grad|: %e\n'],...
           '   ','   ','     ','     ',2*fx, fx,norm_grad);
elseif (verbosity > 1)
   fprintf('************************************************************************\n');
   fprintf(['%3s %3s    k: %5s     num_inner: %5s     %s\n'],...
           '','','______','______','');
   fprintf('       f(x) : %e       |grad| : %e\n',fx,norm_grad);
   fprintf('      Delta : %f\n',Delta);
   if fns.havedist,
      fprintf('       dist : %e\n',stats(end).dist);
   end
%   fprintf('       Time : %f\n',this_time);
end

fail = false;

% **********************
% ** Start of TR loop **
% **********************
inner_it_total = 0;

while stop_outer==0,

   % start clock for this outer iteration
   t0 = tic; % computing time taken per iteration


   % update counter
   k = k+1;

   if (verbosity > 1) || (debug > 0),
      fprintf('************************************************************************\n');
   end

   % *************************
   % ** Begin TR Subproblem **
   % *************************
  
   % determine eta0
   if useRand,
      % random vector in T_x M
      eta = fns.randT(x);
      % must be inside trust-region
      while fns.g(x,eta,eta) > Delta^2,
          %%EDIT%%
          if fns.havescale,
              eta = fns.scale( eta, sqrt(sqrt(eps)) );
          else
              eta = eta * sqrt(sqrt(eps));
          end         
      end
   else
      % without randT, 0*fgradx is the only way that we 
      % know how to create a tangent vector
      if fns.havescale
          eta = fns.scale(fgradx,0);
      else
          eta = 0*fgradx;
      end
   end

   
   
   
   % solve TR subproblem   
   if solver==0
       [eta,numit,stop_inner] = tCG(fns,x,fgradx,eta,Delta,theta,kappa,min_inner,max_inner,useRand,debug);
       srstr = tcg_stop_reason{stop_inner};  
       
   elseif solver==1
       [eta,numit,stop_inner] = tLanczos_rtr(fns,x,fgradx,eta,Delta,theta,kappa,min_inner,max_inner,useRand,debug,epsilon);
       srstr = tcg_stop_reason{stop_inner};
   else
       error('solver must be 0 of 1')
   end
   
   %if stop_inner~=1
   % tinit = exact_search_onlyTxM(fns.prob(), x, eta)
   % [xc_new,fc,succ,numf_a,iarm] = armijo_search(prob, xc,fc,dir, tinit);
   % 
   % %pause
   % eta = scaleTxM(eta,tinit);
   %end
  
   
   
   norm_eta = sqrt(fns.g(x,eta,eta));
      
   % compute the retraction of the proposal
   x_prop  = fns.R(x,eta);
   %[x_prop,fc,shrunk] = fns.shrink_rank(x_prop,1);
   %x_prop.sigma

   % compute function value of the proposal
   fx_prop = fns.f(x_prop); 
     
   % do we accept the proposed solution or not?
   % compute the Hessian at the proposal
   Heta = fns.fhess(x,eta);

   
   % HEURISTIC WARNING:
   % if abs(model change) is relatively zero, we are probably near a critical
   % point. set rho to 1.
%   if abs(rhonum/fx) < em
%      small_rhonum = rhonum;
%      rhonum
%      rhoden
%      rho = 1
%      pause
%   else 
%      small_rhonum = 0;
%   end

   % choose new TR radius based on performance
   trstr = '   ';
   %rhonum
   %rhoden
   
   %rho=1
   
   
   % check the performance of the quadratic model
   rhonum = fx - fx_prop;
   rhoden = -fns.g(x,fgradx,eta) - 0.5*fns.g(x,Heta,eta);
   
   
%    if k>2 && shrunk==false              
%        [x_mag,fx_mag,shrunk] = enlarge_rank(prob, x, opts.nb_k);
%        if fx_mag < fx_prop*(1-1e-8)
%            disp('Choosing enlarged x')
%            %fx_mag
%            %fx_prop
%            %pause
%            
%            rhonum = fx - fx_mag;
%            rhoden = rhoden + (fx_prop - fx_mag);
%            x_prop = x_mag;
%            fx_prop = fx_mag;
%        end
%    end

  

   em = 10*eps;
   dk = em*max(1,abs(fx));
   drhonum = rhonum + dk;
   drhoden = rhoden + dk;
   if abs(rhonum/fx) < 0.1*sqrt(eps) || (abs(drhonum) < em && abs(drhoden) < em)
       rho = 1.0;
       small_rhonum = drhonum;
   else
       rho = drhonum / drhoden;
       small_rhonum = 0;
   end
   
   
   
   if rho < 1/4
      trstr = 'TR-';
      Delta = 1/4*Delta;
   elseif rho > 3/4 && (stop_inner == 2 || stop_inner == 1),
      trstr = 'TR+';
      Delta = min(2*Delta,Delta_bar); 
      %Delta = norm_eta;
   end
   

   % choose new iterate based on performance
   oldgradx = fgradx;
   if rho > rho_prime,
      accept = true;
      accstr = 'acc';
      x    = x_prop;
      fx   = fx_prop;
      fgradx = fns.fgrad(x);
      norm_grad = sqrt(fns.g(x,fgradx,fgradx));
      % precondition the residual
      %if fns.haveprec == 0,
          z = fgradx;
      %else
      %    z = fns.prec(x,fgradx);
      %end
      %crit_measure = sqrt(fns.g(x,fgradx,z)); 
      crit_measure = norm_grad;
      
      %if k>2 %&& shrunk==false          
      %    [x,fx,shrunk] = enlarge_rank(prob, x, opts.nb_k);
      %end
%       fgradx = fns.fgrad(x);
%       norm_grad = sqrt(fns.g(x,fgradx,fgradx));

   else       
      accept = false;
      accstr = 'REJ';
      % line search? -> does not work on diffuclt problems (size 5000, FS
      % 0.01)
      
      
%       accstr = 'LS ';
%       tinit = exact_search_onlyTxM(fns.prob(), x, eta);
%       %tinit = 1;
%       %fx
%       [x,fx,succ,numf_a,iarm,lambda] = armijo_search(fns.prob(), x,fx,eta, tinit);
%       %fx
%       
%       fgradx = fns.fgrad(x);
%       norm_grad = sqrt(fns.g(x,fgradx,fgradx));
%       % precondition the residual
%       %if fns.haveprec == 0,
%           z = fgradx;
%       %else
%       %    z = fns.prec(x,fgradx);
%       %end
%       %crit_measure = sqrt(fns.g(x,fgradx,z)); 
%       crit_measure = norm_grad;
%       
%       %if k>2 %&& shrunk==false          
%       %    [x,fx,shrunk] = enlarge_rank(prob, x, opts.nb_k);
%       %end
%       fgradx = fns.fgrad(x);
%       norm_grad = sqrt(fns.g(x,fgradx,fgradx));
%       norm_eta = abs(lambda)*sqrt(fns.g(x,eta,eta));
%       %if norm_eta < Delta
%       %  trstr = 'TR-';
%       %else
%       %  trstr = 'TR+';
%       %end
%       %Delta = norm_eta;
%       %rho = 0.5;
   end

       
    
    iter_time(k+1) = toc(t0);

   
   % ** Testing for Stop Criteria
%    if norm_grad < opts.abs_grad_tol
%        if verbosity > 0
%            disp('Abs gradient tol reached')
%        end
%         stop_outer = 1;
%         break;
%     end

    if (2*fx) <= opts.tol_cost % for comaparison
        if verbosity > 0
            % disp('Abs f tol reached');
            % fprintf('Cost is %g \n', 2*fx/prob.m);
        end
        stop_outer = 1;
        %break;
    end
%     if sqrt(2*fx)/norm_M_Omega < opts.rel_f_tol
%         if verbosity > 0
%            disp('Rel f tol reached')
%         end
%         stop_outer = 1;
%         break;
%     end
   
  
   %%% EDIT
   if isfield(fns,'normX'),
    norm_x = fns.normX(x);
   else
    norm_x = norm(x.sigma); % TODO function    
   end
   if numit == 0 || norm_eta < 1/100*em*norm_x%*sqrt(max_inner)
       if verbosity > 0
           disp(sprintf('Correction is numerically zero. Quiting at gradient %e',norm_grad))
       end
       stop_outer = 1;
       break
   end

%    % stop clock for this outer step
%    this_time = toc;
%    % update total time
%    total_time = total_time + this_time;

   % ** Record data:
   curstat.k        = k;
   curstat.ng       = norm_grad;
   curstat.crit     = crit_measure;
   curstat.fx       = fx;
   curstat.rho      = rho;
   curstat.rhonum   = rhonum;
   curstat.rhoden   = rhoden;
%   curstat.time     = this_time;
   curstat.accepted = accept;
   curstat.numinner = numit;
   curstat.Delta    = Delta;
   if fns.havedist,
      curstat.dist = fns.dist(x,xsol);
   end
   if useRand,
      curstat.cauchy = used_cauchy;
   end
   stats(end+1) = curstat;
   
  costs(k+1) = fx;
  infos.numit = [infos.numit ; numit];
  infos.hessit = [infos.hessit ; (numit + 1)]; % + 1 because of extra hess computation per outer iteration
  inner_it_total = inner_it_total + numit;
 

   % ** Display:
   if verbosity == 1,
      fprintf(['%3s %3s   k: %5d     num_inner: %5d     ',...
               'cost: % 1.8e   abs.f: % 1.8e   rel.f: %1.8e   |grad|: %1.2e   %s\n'],...
              accstr,trstr,k,numit,2*fx ,sqrt(2*fx),sqrt(2*fx)/norm_M_Omega,norm_grad,srstr);
   elseif verbosity > 1,
      if useRand && used_cauchy,
         fprintf('USED CAUCHY POINT\n');
      end
      fprintf(['%3s %3s    k: %5d     num_inner: %5d     %s\n'],...
              accstr,trstr,k,numit,srstr);
      fprintf('       f(x) : %e     |grad| : %e     |crit| : %e\n',fx,norm_grad,crit_measure);
      fprintf('      Delta : %f          |eta| : %e\n',Delta,norm_eta);
      if small_rhonum ~= 0,
         fprintf('VERY SMALL rho_num: %e\n',small_rhonum);
      else
         fprintf('        rho : %e\n',rho);
      end
      if fns.havedist,
         fprintf('       dist : %e\n',stats(end).dist);
      end
%      fprintf('       Time : %f\n',this_time);
   end
   if debug > 0,
      fprintf('DBG: cos ang(eta,gradf): %d\n',testangle);
      if rho==0
         keyboard;
      end
   end
   
   % stop after max_outer iterations
   if k >= max_outer,
      if (verbosity > 0)
         fprintf('\n*** timed out -- k == %d***\n',k);     
      else if (verbosity > 0),
              fprintf('RTR max its reached,  current gradient is %e, crit measure %e \n',norm_grad,crit_measure);
          end
      end
      fail = true;      
      
      stop_outer = 1;
   end 
     

end  % of TR loop (counter: k)
if (verbosity > 1) || (debug > 0),
   fprintf('************************************************************************\n');
end
% if (verbosity > 0) || (debug > 0)
%     fprintf('Total time is %f\n',total_time);
% end
fprintf('Total inner iterations %i \n', inner_it_total);

infos.iter_time_all = iter_time(1:k+1); 
infos.iter_time = cumsum(iter_time(1:k+1));
infos.costs =  costs(1:k+1);
infos.numit = cumsum(infos.numit);
infos.hessit = cumsum(infos.hessit);

varargout{1} = x;

if nargout == 1,
   varargout{2} = stats;
else
   varargout{2} = stats;
   varargout{3} = fail;
   varargout{4} = infos;
end

end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  get_string %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ret,opts] = get_string(opts,argname,argdesc,def,options)

    % Process inputs and do error-checking 
    errstr = sprintf('%s opts.%s must be: \n',argdesc,argname);
    errstr = [errstr, sprintf('%s  ',options{:})];
    if isfield(opts,argname)
        ret = getfield(opts,argname);
        valid = 0;
        if isstr(ret),
            for i = 1:length(options),
                if isequal(ret,options{i}),
                    valid = 1;
                    break;
                end
            end
        end
        if ~valid,
            error(errstr);
        end

        % remove field from opts
        opts = rmfield(opts,argname);
    else
        ret = def;
    end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  get_int %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ret,opts] = get_int(opts,argname,argdesc,def,lb,ub)

    if nargin < 6
        ub = inf;
    end
    if nargin < 5,
        lb = -inf;
    end
    if nargin < 4,
        mandatory = 1;
        errstr = sprintf('%s opts.%s is mandatory, an integer in [%d,%d]',...
                         argdesc,argname,lb,ub);
    else
        mandatory = 0;
        errstr = sprintf('%s opts.%s must be an integer in [%d,%d]',...
                         argdesc,argname,lb,ub);
    end

    % Process inputs and do error-checking 
    if isfield(opts,argname)
        ret = getfield(opts,argname);
        valid = 0;
        % check that it is an int
        if isnumeric(ret),
            ret = floor(ret);
            % check size (1 by 1) and bounds
            if isequal(size(ret),[1 1]) && lb <= ret && ret <= ub,
                valid = 1;
            end
        end
        if ~valid,
            error(errstr);
        end

        % remove field from opts
        opts = rmfield(opts,argname);
    elseif mandatory == 0,
        ret = def;
    else
        error(errorstr);
    end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%  get_float %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ret,opts] = get_float(opts,argname,argdesc,def,lb,ub)

    if nargin < 6
        ub = inf;
    end
    if nargin < 5,
        lb = -inf;
    end
    if nargin < 4,
        mandatory = 1;
        errstr = sprintf('%s opts.%s is mandatory, a scalar in [%d,%d]',...
                         argdesc,argname,lb,ub);
    else 
        mandatory = 0;
        errstr = sprintf('%s opts.%s must be a scalar in [%d,%d]',...
                         argdesc,argname,lb,ub);
    end

    % Process inputs and do error-checking 
    if isfield(opts,argname)
        ret = getfield(opts,argname);
        valid = 0;
        % check that it is an int
        if isnumeric(ret),
            ret = double(ret);
            if lb <= ret && ret <= ub,
               valid = 1;
            end
        end
        if ~valid,
            error(errstr);
        end

        % remove field from opts
        opts = rmfield(opts,argname);
    elseif mandatory == 0,
        ret = def;
    else
        error(errstr);
    end
end



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%  get_iterate %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [ret,opts] = get_iterate(opts,argname,argdesc,def);

    if nargin < 4,
        mandatory = 1;
        errstr = sprintf('%s opts.%s is mandatory',argdesc,argname);
    else 
        mandatory = 0;
    end

    % Process inputs and do error-checking 
    if isfield(opts,argname)
        ret = getfield(opts,argname);
        % remove field from opts
        opts = rmfield(opts,argname);
    elseif mandatory==0,
        ret = def;
    else
        error(errstr);
    end
end






%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%% Debugging testing of grad/Hess/R %%%%%%%%%%%%%%%%%%%
function TestGH(x,fns)

   % test hessian for symmetry
   for i=1:20,
      T1 = fns.randT(x);
      T2 = fns.randT(x);
      t1ht2 = fns.g( x, T1, fns.fhess(x,T2));
      ht1t2 = fns.g( x, fns.fhess(x,T1), T2);
      fprintf('<T1,H[T2]>: %13e      H[T1],T2: %13e\n',t1ht2,ht1t2);
   end

   fprintf('\n');

   % test that hessian domain is tangent plane
   for i=1:20,
      T = fns.randT(x);
      HT = fns.fhess(x,T);
      PHT = fns.proj(x,HT);
      errnrm = fns.g(x,HT-PHT,HT-PHT);
      errnrm = sqrt(errnrm);
      fprintf('|H[T] - Proj H[T]: %e\n',errnrm);
   end

   fprintf('\n');

   % test model for agreement with fhat
   for i=1:5,
      eta = fns.randT(x);
      eta = eta / sqrt(fns.g(x,eta,eta));
      for t = 10.^[0:-1:-6],
         fxe = fns.f( fns.R(x,t*eta));
         mxe = fns.f( x) + fns.g(x,t*eta,fns.fgrad(x)) + .5*fns.g(x,t*eta,fns.fhess(x,t*eta));
         %f(R_x(t\eta)) - ( f(x) + <gradf(x),t\eta> + 1/2 <\eta,H(x)[\eta]> )
         fprintf('t: %13e     |m-foR|: %13e\n',t,fxe-mxe);
      end
      % move to a new point
      x = fns.R(x,eta);
   end

   fprintf('\n');

   % check that retraction is first-order approximation of the exponential
   if fns.havedist,
      xe = fns.R(x,0*eta);
      fprintf('dist(x, R(x,0)): %13e\n',fns.dist(x,xe));
   end

   fprintf('\n');

   % check that retraction is second-order approximation of the exponential
   % this is not necessary, but may be worth debugging for some retractions
   if fns.havedist,
      for i=1:5,
         eta = fns.randT(x);
         eta = eta / sqrt(fns.g(x,eta,eta));
         for t = 10.^[0:-1:-6],
            xe = fns.R(x,t*eta);
            fprintf('t: %13e     dist(x, R(x,t*eta)): %13e\n',t,fns.dist(x,xe));
         end
         % move to a new point
         x = fns.R(x,eta);
      end
   end
end
