function [xc,histout,costdata,fail] = steep(prob, opts, x0)
%
% Adapted from steep by C. T. Kelley, Dec 20, 1996
%
% function [x,histout,costdata] = steep(x0,f,tol,maxit)
%
% steepest descent with Armijo rule, polynomial linesearch 
% 
%
% Input: x0 = initial iterate
%        f = objective function,
%            the calling sequence for f should be
%            [fout,gout]=f(x) where fout=f(x) is a scalar
%              and gout = grad f(x) is a COLUMN vector
%        tol = termination criterion norm(grad) < tol
%              optional, default = 1.d-6
%        maxit = maximum iterations (optional) default = 1000
%
% Output: x = solution
%         histout = iteration history   
%             Each row of histout is
%            [norm(grad), f, number of step length reductions, iteration count]
%         costdata = [num f, num grad, num hess] (for steep, num hess=0)
%
% Requires: polymod.m
%
% linesearch parms
% 
norm_M_Omega = norm(prob.data);
fail = true;

itc = 1; xc = x0; 
fc = F(prob,xc);
gc = grad(prob,xc);
ip_gc = ip(xc,gc,gc);
% make gc downhill
gc.M = -gc.M; gc.Up = -gc.Up; gc.Vp = -gc.Vp;

numf = 1; numg = 1; numh = 0;
ithist=zeros(opts.maxit,3);
ithist(1,1) = sqrt(ip_gc); ithist(1,2) = fc; ithist(1,3) = 0; 
    
for itc=1:opts.maxit    
    tinit = exact_search_onlyTxM(prob, xc,gc);

    [xc,fc,succ,numf_a,iarm,tmin] = armijo_search(prob, xc, fc, gc, tinit);
    %rank = size(xc_new.V,2)
    %[xc,fc,succ,numf_a,iarm,tmin] = exact_search(prob, fun, xc,fc,gc, 0, 10*tinit);
    %xc = moveEIG(prob, xc, gc, tmin, 1);
    %fc = F(prob, xc);
    %pause
    
    %[xc,fc] = shrink_rank(prob, xc, fc); 
    gc = grad(prob,xc);
    ip_gc = ip(xc,gc,gc);

    ithist(itc,1) = sqrt(ip_gc); ithist(itc,2) = fc; ithist(itc,3) = iarm;
    
    if sqrt(ip_gc) < opts.abs_grad_tol
        if opts.verbosity > 0
            disp('Abs gradient tol reached')
        end
        fail = false;
        break;
    end
    if sqrt(2*fc) < opts.abs_f_tol
        if opts.verbosity > 0
            disp('Abs f tol reached')
        end
        fail = false;
        break;
    end
    if sqrt(2*fc)/norm_M_Omega < opts.rel_f_tol
        if opts.verbosity > 0
            disp('Rel f tol reached')
        end
        fail = false;
        break;
    end
    
    numf = numf + numf_a;
    if ~succ
        histout=ithist(1:itc,:); costdata=[numf, numg, numh];
        x=xc;
        warning('GO:steep',' Line search error steepest descent ')
        return
    end    
        
    % make gc downhill
    gc.M = -gc.M; gc.Up = -gc.Up; gc.Vp = -gc.Vp;
    numg = numg+1;           
end
histout=ithist(1:itc,:); costdata=[numf, numg, numh];
