function [Q, R] = qrup(Q, R, v, u)
% [Q, R] = qrup(Q, R, v, u)
% Rank-one update of the (thin) QR-factorization.
%
% The function computes
%
% [Q, R] = qr(QR + v*u',0)
%
% where Q is a m-by-n orthogonal matrix, R is n-by-n upper triangular,
% v is a m-by-1 vector and u is a n-by-1 vector (m >= n)
%
% Computations are performed in O(mn) arithmetic operations
% instead of O(mn^2) when the factorization is computed from
% stratch, that is, when it is updated using qr(QR + v*u',0).
%
% As opposed to the built-in Matlab function qrupdate,
% this function takes into account the case where Q*R is
% a thin (or compact) QR-factorization, that is, m < n.
%
% Reference for the implemented algorithm: 
%
% J.~W. Daniel, W.~B. Gragg, L. Kaufman and G.~W. Stewart, 
%  ``Reorthogonalization and Stable Algorithms Updating the
%  Gram-Schmidt QR Factorization'', Math. Comp.,  30 (1976), no.
%  136, pp. 772-795.

% Author: Gilles Meyer <g.meyer@ulg.ac.be>
% Last modified: April 2011

    [m, n] = size(Q);
    [mr, nr] = size(R);
    
    nv = length(v);
    nu = length(u);
    
    % Error checking
    if (m < n)
        error('MATLAB:qrup:QWrongOrientation', 'Q must be a vertical or square matrix.')
    elseif (mr ~= nr)
        error('MATLAB:qrup:RNotSquare',...
              'R must be a square upper-triangular matrix.')
    elseif (n ~= mr)
        error('MATLAB:qrup:InnerDimQRfactors',...
            'Inner matrix dimensions of QR factors must agree.')
    elseif (nv ~= m)
        error('MATLAB:qrup:DimUpdateFactor',...
            'Vector v must be of length size(Q,1).')
    elseif (nu ~= n)
        error('MATLAB:qrup:DimUpdateFactor',...
            'Vector u must be of length size(Q,2).')
    end
    
    
    
    % Orthogonalize v wrt. Q
    [y, rho, nre] = reorth_mex(Q, v, norm(v), 1:n, 0.5, 0);
    t = Q'*v;

    %[y, rho, nre, t] = reorth(Q, v);
    
    v = y / rho;

    % Compute Q^R^, where R^ is upper Hessenberg
    [G, tmp] = planerot([t(n);rho]);

    t(n) = tmp(1);
    rho = tmp(2);

    tmp = G*[R(n,n);rho];

    R(n,n) = tmp(1);
    rho = tmp(2);

    tmp = [Q(:,n), v]*G';
    Q(:,n) = tmp(:,1);
    v = tmp(:,2);

    for k = n-1:-1:1,

        p = k:k+1;

        [G, t(p)] = planerot(t(p));

        R(p,k:n) = G*R(p,k:n);
        Q(:,p) = Q(:,p)*G';

    end

    R(1,:) = R(1,:) + t(1) * u';

    
    % Correct for subdiagonal elements in R^
    for k = 1:n-1,

        p = k:k+1;

        [G, R(p,k)] = planerot(R(p,k));

        R(p,k+1:n) = G*R(p,k+1:n);
        Q(:,p) = Q(:,p)*G';

    end

    [G, tmp] = planerot([R(n,n);rho]);
    R(n,n) = tmp(1);
    %rho = tmp(2);

    tmp = [Q(:,n), v]*G';
    Q(:,n) = tmp(:,1);
    %v = tmp(:,2);


end