function [model infos] = svp_(model, rows, cols, b, m, n, k, params)
%Return the result of solving min \| b - X(Omega) \|_F^2 , s.t rank(X) <= k
%Inputs:
%   rows, cols - indices of known entries [b(i) is at position (rows(i),cols(i))] .
%   b - values of known entries
%   k - approximate rank. We only work with matrices of rank up to rank k.
%   params - parameter structure containing the following fields:
%       tol,vtol,mxitr - stopping criteria
%           defaults = 1e-2 (tol), 1e-3 (vtol), 100 (mxitr)
%       verbosity - verbosity level
%           default = 1
%       step_size - step size
%           default = 3/(4p) where p=fraction of entries available
%Outputs:
%   U - Left singular vectors of the optimal matrix X
%   S - Singular values of X
%   V - Right singular vectors of X
%   t - Number of iterations required
%   err - RMSE over the known values
%
%Reference:
% Raghu Meka, Prateek Jain, Inderjit S. Dhillon. 
% "Guaranteed Rank Minimization via Singular Value Projection"
% Arxiv:0909.5457
%Written by: Prateek Jain (pjain@cs.utexas.edu) and Raghu Meka (raghu@cs.utexas.edu)
%Last updated: October, 2009

N = length(b);


if(~exist('rows','var')||~exist('cols','var')||~exist('b','var')||~exist('m','var')||~exist('n','var')||~exist('k','var'))
    error('ErrorSVP:InvalidInputs','Not enough inputs.\n Usage: [U S V t err] = svp (rows,cols, b, m, n, k, params)\n Type "help svp" for more information');
end
if(~isreal(m)||~isreal(n)||~isreal(k))
    error('ErrorSVP:InvalidInputs',...
        'Incorrect input(s). m, n, k should all be real\nUsage: [U S V t err] = svp (rows,cols, b, m, n, k, params)\n Type "help svp" for more information');
end
if(length(cols)~=length(b)||length(rows)~=length(b))
    error('ErrorSVP:InvalidInputs',...
        'Incorrect input. rows/cols and b have equal number of elements. \nUsage: [U S V t err] = svp (rows,cols, b, m, n, k, params)\n Type "help svp" for more information');
end
if(~isempty(find(rows>m, 1))||~isempty(find(cols>n, 1)))
    error('ErrorSVP:InvalidInputs',...
        'Incorrect input. Some elements of rows/cols out of range (i.e., greater than m/n). \nUsage: [U S V t err] = svp (rows,cols, b, m, n, k, params)\n Type "help svp" for more information');
end

p=length(rows)/(m*n); %sampling density

if(~exist('params','var'))
    params=struct();
end
params = SetDefaultParams(params,p);
tol=params.tol;
vtol=params.vtol;
maxtol=params.maxtol;
mxitr=params.mxitr;
verb = params.verb;
stopc = params.stopc;
step_size=params.step_size; %Step Size

% step_type = 'armijo';
step_type = params.step_type;
step_type = 'armijo';

max_step = params.max_step;
sigma_armijo = params.sigma_armijo;


if params.verb,
fprintf('Step = %g\n', step_size);
end


try
  params.decreaselr;
catch
  params.decreaselr = false;
end

normb = norm(b);
nobs  = length(b);

% Compute indices for the known entries
%[I, J] = ind2sub([m,n], Omega);

costs = zeros(mxitr+1,1);

if ~isfield(model,'U') && ~isfield(model,'B') && ~isfield(model,'V'),
  model.U = zeros(m,k);
  model.B = zeros(k,1);
  model.V = zeros(n,k);
  [model.U model.B model.V] = fastsvd(model.U, model.B, model.V, b, rows, cols, k);
else
  model.B = diag(model.B);
end

% Compute value of current iterate X=U*S*V' over the known entries

% X_Omega=proj_omega_sub(model.U*diag(model.B), model.V, rows, cols);
X_Omega = partXY((model.U*diag(model.B))', model.V', rows, cols, N)';

cost = norm(X_Omega - b,'fro')^2/nobs;


costs(1) = cost * nobs;

iter_time = zeros(mxitr+1,1);
iter_time(1) = 0;

ls_iter = zeros(mxitr,1);

t = 1;
t0 = tic;

while  t <= mxitr
  
    if verb,
      fprintf('Iteration\t %d\n',t);
    end

    if verb,
      fprintf('cost: %.5e\n',cost*nobs);
    end
  
    y=(b - X_Omega);
    
    if strcmpi(step_type,'fixed'),
    
      % Take a step in the direction of gradient
      % and compute projection over top k singular vectors
      % i.e, U*S*V'=P_k(U*S*V'+Y)
      [model.U model.B model.V] = fastsvd(model.U, model.B, model.V, step_size*y, rows, cols, k);

%       X_Omega=proj_omega_sub(model.U*diag(model.B), model.V, rows, cols);
      X_Omega=partXY((model.U*diag(model.B))', model.V', rows, cols, N)';


      new_cost = norm(X_Omega - b,'fro')^2/nobs;

    elseif strcmpi(step_type,'decreasing'),
      
      [model.U model.B model.V] = fastsvd(model.U, model.B, model.V, step_size/sqrt(t)*y, rows, cols, k);
      
%       X_Omega=proj_omega_sub(model.U*diag(model.B), model.V, rows, cols);
     X_Omega=partXY((model.U*diag(model.B))', model.V', rows, cols, N)';


      new_cost = norm(X_Omega - b,'fro')^2/nobs;
      
    elseif strcmpi(step_type,'armijo'),
      
      dot_with_grad = norm(y)^2/nobs^2;
%       step_size = max_step;
      step_size = 1/(p*(1 + 1/3));
       
      for j = 1:100,
        
        [new_U, new_B, new_V] = fastsvd(model.U, model.B, model.V, step_size*y, rows, cols, k);
        
%         X_Omega=proj_omega_sub(new_U*diag(new_B), new_V, rows, cols);
        X_Omega=partXY((new_U*diag(new_B))', new_V', rows, cols, N)';


        new_cost = norm(X_Omega - b,'fro')^2/nobs;

        % Check Armijo condition
        armijo = (cost - new_cost) >= (-sigma_armijo * step_size * dot_with_grad);
        if armijo,
          model.U = new_U;
          model.B = new_B;
          model.V = new_V;
          break;
        else
          step_size = step_size / 2;
        end
        ls_iter(t) = j;
        
      end
      
%        if (j == 1), % Adaptive stepsize
%            max_step = 2 * max_step;
%        elseif (j > 2) 
%          max_step = 2 * step_size;     
%        end
       
    end
      
    iter_time(t+1) = toc(t0);
    costs(t+1) = new_cost * nobs;
    cost = new_cost;      
      
%     if mc_stop(costs, t + 1, model, params),
    if new_cost * nobs < params.tol,
      break;
    end

    if new_cost-cost >  maxtol
      error('ErrorSVP:Divergence',['Divergence!!! Current step size: ',num2str(step_size),'. Decrease the step size step_size (params.step_size) and try' ...
  	        ' again']);
    end
    
    t = t+1;
    
end
infos.time = toc(t0);
infos.iter_time = iter_time(1:t+1);
infos.costs = costs(1:t+1);
infos.ls_iter = ls_iter(1:t);
infos.iter = t;
model.iter = t;
infos.cost = infos.costs(end);

function [U,S,V] = fastsvd(U,S,V,y,I,J,k)
%Find the singular values of U S V' + Y
%Y is a sparse matrix
m = size(U,1);
n = size(V,1);
Afunc = @(x) (U*(S.*(V'*x)) + compOmegaYx(m,n,y,I,J,x));
Atfunc = @(x) (V*(S.*(U'*x)) + compOmegaYtx(m,n,y,I,J,x));
[U,Sigma,V] = lansvd(Afunc,Atfunc, m,n,k,'L');
S = diag(Sigma);