% This is a test file that compares algorithms that solve
% "minimize f(X)" with rank(X) = r
%
% Benchmarks are: LMaFit, SVP and LRGeom (embedded approach)
%
% Stopping criterion: cost < tol
%
% Refer "Low-rank optimization with trace norm penalty"
% Technical report (arXiv 1112.2318), 2011 for further details.
% Authors:
% Bamdev Mishra and Gilles Meyer
% {b.mishra, g.meyer}@ulg.ac.be

clear all; close all; clc;
RandStream.setDefaultStream(RandStream('mt19937ar','seed',sum(100*clock)));

cd ..


%% Setup
d1 = 6000;
d2 = 6000;
r = 5;
over_sampling = 4;  
lambda = 0;

dof = (d1 + d2 - r) * r;
f = over_sampling * dof / (d1 * d2);

% Options
tol = 1e-8;
verbosity = false;
random_initialization = false;

fprintf('Rank %i matrix of size %i times %i and OS = %i\n', r, d1, d2, over_sampling);

%% Paramters
params.tol = tol;
params.max_iter_tr = 50;
params.verb = verbosity;

params = default_params(params);


%% Generating random data
data = generate_randn_data(d1, d2, r, f);
 



%% Creating a sparse structure
sparse_structure = sparse(data.ls.rows,data.ls.cols,data.ls.entries,d1,d2);
sparse_structure2 = sparse(data.ls.rows,data.ls.cols,2*data.ls.entries,d1,d2);



%% Setup initial model
model.d1 = d1;
model.d2 = d2;
model.p = r; 
model.sparse_structure = sparse_structure;
model.sparse_structure2 = sparse_structure2;
model.lambda = lambda;



%% Functions
fun_set=@functions_nsym_polar_geometry; % File that specifies geometry of the search space UBV^T
fun_obj=@functions_matrix_completion_polar; % File that computes cost function, gradient of the matrix completion problem



%% Intialization
if ~random_initialization,
    [model.U model.B model.V] = svds(sparse_structure, r);
    fprintf('**** Initialization by taking %i dominant SVD\n', r);
    G = model.U*sqrt(model.B);
    H = model.V*sqrt(model.B);
    
else
    G = randn(model.d1, model.p); H = randn(model.d2, model.p);
    [Qg Rg] = qr(G, 0);
    [Qh Rh] = qr(H, 0);
    [q1 b q2] = svd(Rg * Rh');
    model.U = Qg * q1;
    model.V = Qh * q2;
    model.B = b;
    fprintf('**** Random initialization\n');
end



%%  Trust-region algorithm for fixed rank
fprintf('--------------- Quotient geometry -------------\n');
tic;
 [model_ubv, infos_ubv] = tr_matrix_completion_comparison(data.ts, data.ls, model, params);
toc;



%% Run LMaFit: by Wen et al. 
Known.Ik = data.ls.rows;
Known.Jk = data.ls.cols;
opts.X = G; 
opts.Y = H';
opts.maxit = 500;
opts.tol = tol;
opts.verb = verbosity;
fprintf('--------------- LMaFit -------------\n');

tic;
[X_lma ,Y_lma, infos_lma] = mod_lmafit_mc_adp(d1,d2,Known, data.ls.entries',opts);
toc;




%% Run SVP: by Meka et al., 2009
fprintf('--------------- SVP -------------\n');
model.r = r;
tic;
[model_svp, infos_svp] = svp(data, model, params);
toc;



%% LRGeom RTR by Vandereycken, 2011
fprintf('\n ----------- LRGeom RTR ---------- \n');

M.row = data.ls.rows';
M.col = data.ls.cols';
M.size1 = model.d1;
M.size2 = model.d2;
M.values_Omega = data.ls.entries';

prob.r = r;
prob.k = r;
prob.n1 = M.size1; 
prob.n2 = M.size2;
prob.Omega_i = M.row';
prob.Omega_j = M.col';
prob.data = M.values_Omega'; % we need a column vector!
% prob.Omega = M.ind_Omega';
prob.m = data.ls.nentries;

prob.temp_omega = sparse(prob.Omega_i, prob.Omega_j, prob.data*1,prob.n1,prob.n2,prob.m);
prob.Omega_i = data.ls.rows;
prob.Omega_j = data.ls.cols;

k = r;  
x.V = model.V;
x.sigma = diag(model.B);
x.U = model.U;

par.x = x;
par.opts.tol_cost = tol;
par.opts.verbosity = +verbosity;
par.rank = r;
par.maxiter = 50;
par.prob = prob;


[X_lrgeom,out_lrgeom] = geom_rtr_common(M, par);


%% Plots

% Cost vs. iterations
figure;
fs = 20;
semilogy((infos_ubv.costs),'--+','Color','b','LineWidth',2); hold on;
semilogy((infos_lma.costs),'--o','Color','r','LineWidth',2); 
semilogy(infos_svp.costs,'Color',[0.7 0.5 0],'LineStyle','--','Marker','p','LineWidth',2); 
semilogy((X_lrgeom.infos.costs),'k--s','Linewidth',2);
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel('Number of iterations'); ylabel('Cost in log scale');
legend('Trust-region','LMaFit', 'SVP','LRGeom');
legend 'boxoff';
title([num2str(d1),' by ',num2str(d2),', rank ',num2str(r),', OS is ',num2str(over_sampling)])
box off;


% Cost vs time
figure;
fs = 20;
semilogy(infos_ubv.iter_time,(infos_ubv.costs),'--+','Color','b','LineWidth',2); hold on; 
semilogy(infos_lma.iter_time,(infos_lma.costs),'-o','Color','r','LineWidth',2); 
semilogy(infos_svp.iter_time,infos_svp.costs,'Color',[0.7 0.5 0],'LineStyle','--','Marker','p','LineWidth',2); 
semilogy(X_lrgeom.infos.iter_time,(X_lrgeom.infos.costs),'k--s','LineWidth',2);
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel('Time taken (sec.)'); ylabel('Cost in log scale');
legend('Trust-region','LMaFit', 'SVP', 'LRGeom');
legend 'boxoff';
title([num2str(d1),' by ',num2str(d2),', rank ',num2str(r),', OS is ',num2str(over_sampling)])
box off;


