% This is a test file that solves
% a problem of the type "minimize f(X) + \lambda ||X||_*" 
% for a fixed lambda with a sequence of lambda values that approaches to
% the final one. The idea is similar to the continuation method.
%
% We create a set of local tolerances.
%
% Stopping criterion: difference in consecuitve costs < tol || rel. difference in cost < tol
%
% Benchmarks are: NNLS
%
% Refer "Low-rank optimization with trace norm penalty"
% Technical report (arXiv 1112.2318), 2011 for further details.
%
% Authors:
% Bamdev Mishra <b.mishra@ulg.ac.be>

clear all; close all; clc;
RandStream.setDefaultStream(RandStream('mt19937ar', 'seed', sum(100*clock)));

cd ..



%% Setup
d1 = 1000;
d2 = 1000;
r = 5;
over_sampling = 5;  


dof = (d1 + d2 - r) * r;
f = over_sampling * dof / (d1 * d2);
fprintf('Rank %i matrix of size %i times %i and OS = %i\n', r, d1, d2, over_sampling);

% Options
verbosity = false;
tol = 1e-12;

%% Optimization algorithm is the trust-region algorithm
optim_algo = @tr_matrix_completion_regpath_comparison;




%% Paramters
params.tol = tol; % Global tolerances for target lambda
params.vtol = tol;
params.grad_tol = 1e-5;
params.dg_tol = 1e-3;
params.dg_vtol = 1e-5;
params.smax_tol = 1e-3;
params.max_iter = 10000;
params.max_iter_tr = 50;
params.verb = verbosity; % Show more output

params = default_params(params);


params.local_tol = 1e-5; % Relaxed tolerances
params.local_vtol = 1e-5;
params.local_grad_tol = 1e0;
params.local_dg_tol = 1;
params.local_dg_vtol = 1;



%% Generating random data
data = generate_randn_data(d1, d2, r, f);
 

%% Creating a sparse structure
sparse_structure = sparse(data.ls.rows,data.ls.cols,data.ls.entries,d1,d2);
sparse_structure2 = sparse(data.ls.rows,data.ls.cols,2*data.ls.entries,d1,d2);



%% Lambda parameter
lambda_threshold = 2*lansvd(sparse_structure, 1);
lambda0 = lambda_threshold;% should be less than lambda_threshold for rank to be non-zero
lambda = 1e-5*lambda0; % Similar to ones in NNLS




%% Setup initial model
model.d1 = d1;
model.d2 = d2;

model.p = 1; 
model.pmax = min(d1, d2);
model.sparse_structure = sparse_structure;
model.sparse_structure2 = sparse_structure2;



%% Functions
fun_set=@functions_nsym_polar_geometry; % File that specifies geometry of the search space UBV^T
fun_obj=@functions_matrix_completion_polar; % File that computes cost function, gradient of the matrix completion problem




%% Random Intialization
G = randn(model.d1, model.p); H = randn(model.d2, model.p);
[Qg Rg] = qr(G, 0);
[Qh Rh] = qr(H, 0);
[q1 b q2] = svd(Rg * Rh');
model.U = Qg * q1;
model.V = Qh * q2;
model.B = b;



%% Run algorithm
fprintf('------------------ Quotient geometry ------------------\n')


lambda_array = generate_a_sequence(lambda0 / 2, lambda);
% lambda_array = [lambda0/2, lambda0/4, lambda]; % Any given sequence of parameters

fprintf('NOTE! \n')
fprintf('We propose a continuation method by creating a grid of lambda values.\n')
fprintf('The stopping criterion is relaxed for all except the target lambda.\n')
fprintf('Target lambda is %.5e.\n',lambda);
fprintf('We traverse through...\n');
for kk = 1: length(lambda_array); 
   fprintf('%s \n', num2str(lambda_array(kk), '%10.5e') ); 
end
fprintf('\n\n');

model.lambda_array = lambda_array;
tic;
[model_ubv, infos_ubv] = compute_path_comparison(optim_algo, data, model, params);
toc;
infos_ubv.iter_time(1) = 0;


%% NNLS (K. C. Toh)
fprintf('\n\n');
fprintf('------------------ APGL: Acc. Proximal Gradient method ------------------\n')

II = data.ls.rows';
JJ = data.ls.cols';
bb = data.ls.entries;
Jcol = compJcol(JJ); 
nr = d1;
nc = d2;

problem_type = 'NNLS'; 


options.tol = 1e-8;
mumax = svds(sparse(data.ls.rows,data.ls.cols,data.ls.entries,d1,d2),1,'L',options);  % lambda_threshol / 2 !
mu_scaling = 1e-4;
mutarget = lambda / 2; % Note !
noiseratio = 0;

Amap  = @(X) Amap_MatComp(X,II,Jcol);
if (exist('mexspconvert')==3);
    ATmap = @(y) mexspconvert(nr,nc,y,II,Jcol);
else
    ATmap = @(y) spconvert([II,JJ,y; nr,nc,0]);
end
par.tol     = tol;
par.verbose = 1;
par.truncation_start = 10; % As suggested by the authors
par.continuation_scaling = mu_scaling;
par.continuation = true; % Continuation is efficient; Without continuation, the algorithm performs (very) poorly.
par.linesearch = true; % This makes the algorithm efficient
par.truncation = true; % This makes the algorithm efficient as normally iterates are not low-rank and hence, need to be truncated.
par.maxiter = 300;
par.X_old.U = G;
par.X_old.V = H;
par.verbose = verbosity; % Show output

if par.continuation,
    fprintf('YES continuation, ');
else
    fprintf('NO continuation!, ');
end

if par.linesearch,
    fprintf('YES linesearch, ');
else
    fprintf('NO linesearch!, ');
end
if par.truncation,
    fprintf('YES truncation...\n\n');
else
    fprintf('NO truncation!\n\n');
end


tic;
[X, iter, time, sd, hist] = mod_APGL(nr,nc,problem_type,Amap,ATmap,bb,mutarget,0,par);
toc;





%% Plots

% Cost vs. iterations
figure;
fs = 20;
plot(log10(infos_ubv.costs),'--+','Color','b','LineWidth',2); hold on;
plot(log10(hist.costs),'--o','Color',[0.5 0.3 0.3],'LineWidth',2); 
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel('Number of iterations'); ylabel('Cost in log scale');
legend('Descent-restart + TR','APGL');
legend 'boxoff';
title(['Continuation used: ', num2str(d1),' by ',num2str(d2),', rank ',num2str(r),', OS is ',num2str(over_sampling),', \lambda = ', num2str(lambda)])
box off;



% Cost vs time
figure;
fs = 20;
plot(infos_ubv.iter_time, log10(infos_ubv.costs),'--+','Color','b','LineWidth',2); hold on; 
plot(hist.iter_time, log10(hist.costs),'--o','Color',[0.5 0.3 0.3],'LineWidth',2); 
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel('Time taken (sec.)'); ylabel('Cost in log scale');
legend('Descent-restart + TR','APGL');
legend 'boxoff';
title(['Continuation used: ', num2str(d1),' by ',num2str(d2),', rank ',num2str(r),', OS is ',num2str(over_sampling),', \lambda = ', num2str(lambda)])
box off;